import numpy as np
from pathlib import Path

import jax

import pytest

import nqxpack
import netket as nk
from nqxpack._src import distributed

hi = nk.hilbert.Spin(0.5, 4)
g = nk.graph.Chain(4)

operators = {}
operators["IsingNumba"] = nk.operator.IsingNumba(hi, h=1.0, graph=g)
operators["IsingJax"] = nk.operator.IsingJax(hi, h=1.0, graph=g)
operators["LocalOperatorNumba"] = operators["IsingNumba"].to_local_operator()
operators["LocalOperatorJax"] = operators["IsingJax"].to_local_operator()
operators["PauliStringsNumba"] = operators["LocalOperatorNumba"].to_pauli_strings()
operators["PauliStringsJax"] = operators["LocalOperatorJax"].to_pauli_strings()

operators_params = [pytest.param(op, id=name) for name, op in operators.items()]


@pytest.mark.parametrize("operator", operators_params)
def test_save_mcstate(operator, tmpdir):
    if distributed.mode() == "sharding":
        tmpdir = Path(distributed.broadcast_string(str(tmpdir)))

    nqxpack.save(operator, tmpdir / "operator.mpack")
    distributed.barrier("barrier 1")
    loaded_operator = nqxpack.load(tmpdir / "operator.mpack")

    assert operator.hilbert == loaded_operator.hilbert
    if not isinstance(operator, nk.operator.DiscreteJaxOperator):
        operator = operator.to_jax_operator()
        loaded_operator = loaded_operator.to_jax_operator()

    op_flat, treestruct = jax.tree.flatten(operator)
    loaded_op_flat, loaded_treestruct = jax.tree.flatten(loaded_operator)
    for a, b in zip(op_flat, loaded_op_flat):
        np.testing.assert_allclose(a, b)
    assert treestruct == loaded_treestruct



__all__ = [
    "register_serialization",
    "register_automatic_serialization",
    "register_closure_simple_serialization",
    "AssetManager",
]

from nqxpack._src.lib_v1.custom_types import (
    register_serialization as register_serialization,
    register_automatic_serialization as register_automatic_serialization,
)
from nqxpack._src.lib_v1.closure import (
    register_closure_simple_serialization as register_closure_simple_serialization,
)
from nqxpack._src.lib_v1.asset_lib import AssetManager as AssetManager



from typing import Callable, Any
from nqxpack._src.contextmgr import current_context


class VersionedDeserializationRegistry:
    """
    Registry for versioned deserialization functions.

    Stores deserialization functions keyed by class path strings, with support
    for multiple versions of the same class.

    Each entry maps a class path string to a list of tuples containing:
    - minimum version (inclusive) as a tuple (major, minor, patch)
    - deserialization function

    When deserializing, the registry selects the appropriate function based on
    the package version stored in the context.
    """

    def __init__(self):
        # Dict[str, List[Tuple[Tuple[int, int, int], Callable]]]
        self._registry = {}

    def register(
        self,
        class_path: str,
        deserialization_fun: Callable[[dict], Any],
        min_version: tuple[int, int, int] = (0, 0, 0),
    ):
        """
        Register a deserialization function for a specific class path and version range.

        Args:
            class_path: The fully qualified class path as a string (e.g., "package.module.Class")
            deserialization_fun: Function that takes a dict and returns an instance
            min_version: Minimum version (inclusive) for which this deserializer is valid
        """
        if class_path not in self._registry:
            self._registry[class_path] = []

        # Insert in sorted order by version (highest to lowest)
        entry = (min_version, deserialization_fun)
        entries = self._registry[class_path]

        # Find insertion point - keep sorted by version descending
        insert_idx = 0
        for i, (existing_version, _) in enumerate(entries):
            if min_version >= existing_version:
                insert_idx = i
                break
            insert_idx = i + 1

        entries.insert(insert_idx, entry)

    def get(self, class_path: str, package_name: str = None) -> Callable | None:
        """
        Get the appropriate deserialization function for a class path.

        Args:
            class_path: The fully qualified class path as a string
            package_name: Name of the package to check version for. If None,
                         extracts from class_path

        Returns:
            The deserialization function, or None if not found
        """
        if class_path not in self._registry:
            return None

        # Extract package name from class path if not provided
        if package_name is None:
            # Assume format "package.module.Class"
            parts = class_path.split(".")
            if len(parts) > 0:
                package_name = parts[0]
            else:
                package_name = None

        # Get the version from the saved file metadata
        # The context manager already returns parsed tuples
        ctx = current_context()
        try:
            saved_versions = ctx.saved_file_package_versions
            file_version = saved_versions.get(package_name, (0, 0, 0))
        except (AttributeError, KeyError, TypeError):
            # If no version info available, use (0, 0, 0)
            file_version = (0, 0, 0)

        # Find the appropriate deserializer
        # Entries are sorted by version descending, so we find the first one
        # where min_version <= file_version
        for min_version, deserialization_fun in self._registry[class_path]:
            if file_version >= min_version:
                return deserialization_fun

        # No suitable version found
        return None

    def __contains__(self, class_path: str) -> bool:
        """Check if a class path is registered."""
        return class_path in self._registry


# Global instance
VERSIONED_DESERIALIZATION_REGISTRY = VersionedDeserializationRegistry()



import numpy as np
from pathlib import Path

import jax
import jax.numpy as jnp
from flax import nnx
from flax import linen

import pytest

import nqxpack
from nqxpack._src import distributed
import netket as nk

hi = nk.hilbert.Spin(0.5, 4)
g = nk.graph.Chain(4)

models = {}
models["RBM"] = nk.models.RBM(param_dtype=complex)
models["SymmSymExp"] = nk.nn.blocks.SymmExpSum(
    nk.models.RBM(param_dtype=complex), symm_group=g.translation_group()
)


def build_nnx_model():
    return nnx.Sequential(
        nnx.Linear(
            in_features=4,
            out_features=2,
            rngs=nnx.Rngs(1),
        ),
        # nnx.relu,
        nnx.Linear(
            in_features=2,
            out_features=1,
            rngs=nnx.Rngs(1),
        ),
        jnp.squeeze,
    )


models["NNX-Sequential"] = build_nnx_model

models_params = [pytest.param(model, id=name) for name, model in models.items()]


@pytest.mark.parametrize("model", models_params)
def test_save_mcstate(model, tmpdir):
    if distributed.mode() == "sharding":
        tmpdir = Path(distributed.broadcast_string(str(tmpdir)))

    if not isinstance(model, linen.Module):
        model = model()
    sa = nk.sampler.MetropolisLocal(hi, n_chains=4)
    vs = nk.vqs.MCState(sa, model, n_samples=64)

    distributed.barrier("barrier 1")
    nqxpack.save(vs, tmpdir / "mcstate.mpack")
    nqxpack.save(vs, "mcstate.mpack")
    distributed.barrier("barrier 2")

    new_vs = nqxpack.load(tmpdir / "mcstate.mpack")

    assert vs.hilbert == new_vs.hilbert
    jax.tree.map(np.testing.assert_allclose, vs.parameters, new_vs.parameters)
    jax.tree.map(
        np.testing.assert_allclose,
        distributed.allgather(vs.samples),
        distributed.allgather(new_vs.samples),
    )



from nqxpack._src.lib_v1.closure import register_closure_simple_serialization
from nqxpack._src.lib_v1.custom_types import (
    register_serialization,
    register_deserialization,
)

import jax

register_closure_simple_serialization(
    jax.nn.initializers.normal,
    "init",
    original_qualname="jax._src.nn.initializers.normal",
)
register_closure_simple_serialization(
    jax.nn.initializers.variance_scaling,
    "init",
    original_qualname="jax._src.nn.initializers.variance_scaling",
)


def serialize_PyTreeDef(obj):
    return {
        "node_data": obj.node_data(),
        "children": obj.children(),
    }


def deserialize_PyTreeDef(obj):
    # Fall back to old method if available (for older JAX versions)
    if hasattr(jax.tree_util.PyTreeDef, "make_from_node_data_and_children"):
        return jax.tree_util.PyTreeDef.make_from_node_data_and_children(
            jax.tree_util.default_registry, obj["node_data"], obj["children"]
        )

    # Workaround for new JAX versions with old serialized data
    # We need to reconstruct the PyTreeDef from node_data and children
    return _reconstruct_pytreedef_from_legacy(obj)


def _reconstruct_pytreedef_from_legacy(obj):
    """
    Reconstruct a PyTreeDef from the legacy serialization format.

    This function implements the logic from the old make_from_node_data_and_children
    method that was removed from JAX in https://github.com/jax-ml/jax/commit/0e7c96a54a98d0f2b7e1c29bd7cc61b1b9bcbf59
    """
    node_data, children = obj["node_data"], obj["children"]
    import jax.tree_util as tu

    # If we have a leaf node
    if node_data is None:
        # Create a PyTreeDef for a leaf
        _, leaf_treedef = tu.tree_flatten(object())
        return leaf_treedef

    # node_data is a tuple of (node_type, metadata)
    node_type, metadata = node_data

    # First, recursively reconstruct all children PyTreeDefs
    reconstructed_children = []
    for child in children:
        if isinstance(child, dict) and "node_data" in child and "children" in child:
            # This is a serialized PyTreeDef that needs reconstruction
            child_treedef = _reconstruct_pytreedef_from_legacy(
                child["node_data"], child["children"]
            )
            reconstructed_children.append(child_treedef)
        else:
            # This is already a PyTreeDef
            reconstructed_children.append(child)

    # Now we need to build a template structure and extract dummy leaf values
    # to create properly structured nested data

    # Create dummy data for each child based on their treedefs
    child_leaves_list = []
    for child_treedef in reconstructed_children:
        num_leaves = child_treedef.num_leaves  # This is a property, not a method
        # Create dummy leaves for this child
        dummy_leaves = [object() for _ in range(num_leaves)]
        # Unflatten to get the properly structured child
        child_structure = tu.tree_unflatten(child_treedef, dummy_leaves)
        child_leaves_list.append(child_structure)

    # Build a template structure based on the node type and metadata
    if node_type.__name__ == "tuple" and hasattr(node_type, "_fields"):
        # NamedTuple
        template = node_type(*child_leaves_list)
    elif node_type is tuple:
        # Regular tuple
        template = tuple(child_leaves_list)
    elif node_type is list:
        # List
        template = list(child_leaves_list)
    elif node_type is dict:
        # Dict - metadata contains the keys
        if metadata is not None:
            template = {key: child for key, child in zip(metadata, child_leaves_list)}
        else:
            template = {}
    else:
        # Custom pytree type
        # Most custom pytree types can be created from their children
        template = tuple(child_leaves_list)

    # Get the treedef from the template
    _, template_treedef = tu.tree_flatten(template)

    return template_treedef


register_serialization(
    jax.tree_util.PyTreeDef, serialize_PyTreeDef, deserialize_PyTreeDef
)

# Register versioned deserialization for the old jaxlib path
# In older JAX versions, PyTreeDef was exposed as jaxlib.xla_extension.pytree.PyTreeDef
# Now it's jax.tree_util.PyTreeDef
register_deserialization(
    "jaxlib.xla_extension.pytree.PyTreeDef",
    _reconstruct_pytreedef_from_legacy,
    min_version=(0, 0, 0),
)



from abc import ABC, abstractmethod
from pathlib import Path

from zipfile import ZipFile

import jax
from flax import serialization

from nqxpack._src.contextmgr import current_context


class AssetManager(ABC):
    """
    Used to store binary blobs or large serialized objects, which may be needed
    when serializing some custom types.

    This is passed in as a parameter to the `serialize` and `deserialize` functions.
    """

    @abstractmethod
    def _write(self, key: str, value: bytes):
        """
        Commits a binary blob to the asset manager under key `key`.

        This should be implemented for the specific asset manager.
        """
        pass

    @abstractmethod
    def _read(self, key: str) -> bytes:
        """
        Reads a binary blob to the asset manager under key `key`.

        This should be implemented for the specific asset manager.
        """
        pass

    def write_asset(self, asset_name, value: bytes, path: tuple[str, ...] = None):
        """
        Write an asset to the backend

        Args:
            asset_name: Name of the asset
            value: The asset to write, which should be a binary blob.
            path: Path to the asset, as a tuple of strings. This is optional.
        """
        if path is None:
            path = current_context().path
        key = f"{path}/{asset_name}"
        return self._write(key, value)

    def read_asset(self, asset_name, path: tuple[str, ...] = None):
        if path is None:
            path = current_context().path
        key = f"{path}/{asset_name}"
        return self._read(key)

    def write_msgpack(self, asset_name, value: dict, path: tuple[str, ...] = None):
        """
        Write a dictionary of msgpack-serializable objects to the asset manager.

        Args:
            asset_name: Name of the asset
            value: The asset to write, which should be a dictionary of msgpack-serializable objects.
            path: Path to the asset, as a tuple of strings. This is optional.
        """
        if jax.process_index() == 0:
            self.write_asset(asset_name, serialization.msgpack_serialize(value), path)

    def read_msgpack(self, asset_name, path: tuple[str, ...] = None):
        """
        Reads a dictionary of data serialized with msgpack to the asset manager.

        Args:
            asset_name: Name of the asset
            value: The asset to write, which should be a dictionary of msgpack-serializable objects.
            path: Path to the asset, as a tuple of strings. This is optional.
        """
        return serialization.msgpack_restore(self.read_asset(asset_name, path))


class InMemoryAssetManager(AssetManager):
    """
    Asset manager that stores assets in memory.
    """

    def __init__(self):
        self._assets = {}

    def _write(self, key: str, value: bytes):
        self._assets[key] = value

    def _read(self, key: str) -> bytes:
        return self._assets[key]


class FolderAssetManager(AssetManager):
    def __init__(self, folder, path, remove_root=None):
        """
        Constructs an asset manager backed by a folder.

        Args:
            folder: a directory to store the asset.
            root: A prefix to remove from all keys when writing to the archive. This is optional.
        """
        if not isinstance(folder, Path):
            folder = Path(folder)
        self.folder = folder
        self.path = path
        self.remove_root = remove_root

    def _write(self, key: str, value: bytes):
        if self.remove_root is not None:
            if key.startswith(self.remove_root):
                key = key[len(self.remove_root) :]
        if self.path is not None:
            key = self.path + key

        if jax.process_index() == 0:
            if not (self.folder / key).parent.exists():
                (self.folder / key).parent.mkdir(parents=True)
            with open(self.folder / key, "wb") as f:
                f.write(value)

    def _read(self, key: str) -> bytes:
        if self.remove_root is not None:
            if key.startswith(self.remove_root):
                key = key[len(self.remove_root) :]
        if self.path is not None:
            key = self.path + key

        with open(self.folder / key, "rb") as f:
            return f.read()


class ArchiveAssetManager(AssetManager):
    """
    Asset manager that writes to a zip archive.
    """

    def __init__(
        self, archive: ZipFile, path: str | None = None, remove_root: str | None = None
    ):
        """
        Constructs an asset manager backed by a zip file archive.

        Args:
            archive: an open zip file object.
            remove_root: A prefix to remove from all keys when writing to the archive. This is optional.
            path: A prefix to add to all keys when writing to the archive. This is optional.
        """
        self.archive = archive
        self.path = path
        self.remove_root = remove_root

    def _write(self, key: str, value: bytes):
        if self.remove_root is not None:
            if key.startswith(self.remove_root):
                key = key[len(self.remove_root) :]
        if self.path is not None:
            key = self.path + key
        if jax.process_index() == 0:
            with self.archive.open(key, "w") as f:
                f.write(value)

    def _read(self, key: str) -> bytes:
        if self.remove_root is not None:
            if key.startswith(self.remove_root):
                key = key[len(self.remove_root) :]
        if self.path is not None:
            key = self.path + key

        if key not in self.archive.namelist():
            raise FileNotFoundError(f"Asset {key} not found in archive.")

        with self.archive.open(key, "r") as f:
            return f.read()



from nqxpack._src.lib_v1 import (
    serialize_object,
    deserialize_object,
)

from omegaconf import DictConfig, ListConfig

from .. import common


@common.skipif_distributed
def test_list_and_dict_conversion():
    d = {"a": 1, "b": [1, 2, 3]}
    d_cfg = DictConfig(d)
    l = [1.0, 2.0, 3.0]
    l_cfg = ListConfig(l)

    assert serialize_object(l) == serialize_object(l_cfg)
    assert serialize_object(d) == serialize_object(d_cfg)
    assert deserialize_object(serialize_object(l)) == l
    assert deserialize_object(serialize_object(d)) == d



import jax

import flax
from flax import nnx
from flax import serialization

from nqxpack._src.lib_v1.custom_types import (
    register_automatic_serialization,
)
from nqxpack._src.lib_v1 import register_serialization
from nqxpack._src.contextmgr import current_context


# Graph

# flax.nnx
from flax.nnx.graph import HashableMapping

register_automatic_serialization(HashableMapping, mapping="_mapping")

flax_version = tuple(int(x) for x in flax.__version__.split("."))[:3]
if flax_version >= (0, 10, 4):
    from flax.nnx.graph import IndexesPytreeDef

    register_automatic_serialization(IndexesPytreeDef, *IndexesPytreeDef._fields)

# Sequence
register_automatic_serialization(nnx.Sequential, _args_="layers")


def serialize_linear(
    layer: nnx.Linear,
) -> dict:
    asset_manager = current_context().asset_manager
    _, data = nnx.split(layer)
    state_dict = serialization.to_state_dict(data.to_pure_dict())

    asset_manager.write_msgpack("state.msgpack", state_dict)

    return {
        "in_features": layer.in_features,
        "out_features": layer.out_features,
        "use_bias": layer.use_bias,
        "dtype": layer.dtype,
        "param_dtype": layer.param_dtype,
        "precision": layer.precision,
        "kernel_init": layer.kernel_init,
        "bias_init": layer.bias_init,
        "dot_general": layer.dot_general,
        "promote_dtype": layer.promote_dtype,
        # "preferred_element_type": layer.preferred_element_type,
    }


def deserialize_linear(obj: dict) -> nnx.Linear:
    asset_manager = current_context().asset_manager

    layer = nnx.Linear(
        obj["in_features"],
        obj["out_features"],
        use_bias=obj["use_bias"],
        dtype=obj.get("dtype", jax.numpy.float32),
        param_dtype=obj.get("param_dtype", jax.numpy.float32),
        precision=obj.get("precision", None),
        kernel_init=obj.get("kernel_init", nnx.initializers.lecun_normal()),
        bias_init=obj.get("bias_init", nnx.initializers.zeros),
        dot_general=obj.get("dot_general", None),
        promote_dtype=obj.get("promote_dtype", False),
        # preferred_element_type=obj.get("preferred_element_type", None),
        rngs=nnx.Rngs(0),
    )
    graph, data = nnx.split(layer)
    state_dict = asset_manager.read_msgpack("state.msgpack")
    state = serialization.from_state_dict(data.to_pure_dict(), state_dict)
    return nnx.merge(graph, state)


register_serialization(
    nnx.Linear, serialize_linear, deserialization_fun=deserialize_linear
)



from functools import partial


# flake8: noqa: E402
from nqxpack._src.lib_v1.custom_types import (
    register_serialization,
    register_automatic_serialization,
)
from nqxpack._src.contextmgr import current_context
from nqxpack._src.distributed import replicate_sharding

import jax
from flax import serialization

# Graph
from netket.graph import Lattice
from netket.utils.version_check import module_version


def serialize_Lattice(g):
    return {
        "basis_vectors": g.basis_vectors,
        "extent": g.extent.tolist(),
        "pbc": g.pbc.tolist(),
        "site_offsets": g._site_offsets,
        "point_group": g._point_group,
        "max_neighbor_order": g._max_neighbor_order,
    }


register_serialization(Lattice, serialize_Lattice)

# Hilbert
from netket.hilbert import Spin, Qubit, Fock
from netket.hilbert import SpinOrbitalFermions
from netket.hilbert.constraint import ExtraConstraint


def serialize_Spin(hi):
    return {
        "s": hi._s,
        "N": hi.size,
        "total_sz": hi._total_sz,
        "inverted_ordering": hi._inverted_ordering,
        "constraint": hi.constraint if hi._total_sz is None else None,
    }


register_serialization(Spin, serialize_Spin)

register_automatic_serialization(Qubit, N="size")
register_automatic_serialization(
    Fock,
    "n_max",
    "n_particles",
    N="size",  # constraint="constraint"
)


def serialize_SpinOrbitalFermions(hi):
    data = {
        "n_orbitals": hi.n_orbitals,
        "s": hi.spin,
    }
    constraint = hi.constraint
    if hi.spin is None:
        data["n_fermions"] = hi.n_fermions
    elif any(s is not None for s in hi.n_fermions_per_spin):
        data["n_fermions_per_spin"] = hi.n_fermions_per_spin

    # Set the constraint as None for the default constraint
    if any(s is not None for s in hi.n_fermions_per_spin):
        if isinstance(constraint, ExtraConstraint):
            constraint = constraint.extra_constraint
        else:
            constraint = None
    data["constraint"] = constraint
    return data


register_serialization(SpinOrbitalFermions, serialize_SpinOrbitalFermions)

from netket.hilbert import DoubledHilbert


def serialize_DoubledHilbert(hi):
    return {
        "hilb": hi.physical,
    }


def deserialize_DoubledHilbert(obj):
    return DoubledHilbert(obj["hilb"])


register_serialization(
    DoubledHilbert,
    serialize_DoubledHilbert,
    deserialization_fun=deserialize_DoubledHilbert,
)

from netket.hilbert.tensor_hilbert import TensorHilbert
from netket.hilbert.tensor_hilbert_discrete import TensorDiscreteHilbert


def serialize_TensorHilbert(hi):
    return {"_args_": hi._hilbert_spaces}


register_serialization(
    TensorHilbert,
    serialize_TensorHilbert,
)
register_serialization(
    TensorDiscreteHilbert,
    serialize_TensorHilbert,
)


# Constraints
from netket.hilbert.constraint import SumConstraint, SumOnPartitionConstraint

register_automatic_serialization(SumConstraint, "sum_value")
register_automatic_serialization(SumOnPartitionConstraint, "sum_values", "sizes")
register_automatic_serialization(ExtraConstraint, "base_constraint", "extra_constraint")

# Sampler
from netket.sampler import MetropolisSampler, ExactSampler

register_automatic_serialization(
    MetropolisSampler,
    "hilbert",
    "rule",
    "sweep_size",
    "reset_chains",
    "n_chains",
    "chunk_size",
    "machine_pow",
    "dtype",
    array_to_list=True,
)
register_automatic_serialization(
    ExactSampler, "hilbert", "machine_pow", "dtype", array_to_list=True
)

# Sampler Rules
from netket.sampler.rules import (
    ExchangeRule,
    FixedRule,
    HamiltonianRule,
    LocalRule,
    MultipleRules,
    TensorRule,
)
from netket.sampler import rules as sampler_rules

if hasattr(sampler_rules, "GlobalSpinFlipRule"):
    register_automatic_serialization(sampler_rules.GlobalSpinFlipRule)

register_automatic_serialization(FixedRule)
register_automatic_serialization(LocalRule)
register_automatic_serialization(ExchangeRule, "clusters", array_to_list=True)
register_automatic_serialization(
    MultipleRules, "rules", "probabilities", array_to_list=True
)
register_automatic_serialization(TensorRule, "hilbert", "rules")
register_automatic_serialization(HamiltonianRule, "operator")


# group theory
from netket.graph.space_group import Translation, Permutation, TranslationGroup


def serialize_translation(t):
    if module_version("netket") >= (3, 18, 0):
        return {
            "inverse_permutation_array": t.inverse_permutation_array.tolist(),
            "displacement": t._vector.tolist(),
        }
    else:
        return {
            "permutation": t.permutation.wrapped.tolist(),
            "displacement": t._vector.tolist(),
        }


register_serialization(Translation, serialize_translation)


def serialize_permutation(t):
    if module_version("netket") >= (3, 18, 0):
        return {
            "inverse_permutation_array": t.inverse_permutation_array.tolist(),
            "name": t._name,
        }
    else:
        return {
            "permutation": t.permutation.wrapped.tolist(),
            "name": t._name,
        }


register_serialization(Permutation, serialize_permutation)


# groups
def serialize_translationgroup(tg):
    return {"lattice": tg.lattice, "axes": tg.axes}


def deserialize_translationgroup(obj):
    if "elems" in obj:
        # this was serialized with an old version of ketnet
        from netket.graph.space_group import PermutationGroup

        return PermutationGroup(**obj)
    else:
        return TranslationGroup(**obj)


register_serialization(
    TranslationGroup,
    serialize_translationgroup,
    deserialization_fun=deserialize_translationgroup,
)

try:
    from netket.utils.model_frameworks.nnx import NNXWrapper
except ModuleNotFoundError:
    raise ImportError(
        "This version of netket pro requires a more recent netket version. Update NETKET! (from github)"
    )

register_automatic_serialization(NNXWrapper, "graphdef")

# mcstate
from netket.vqs import MCMixedState, MCState, FullSumState


def _replicate(x):
    if isinstance(x, jax.Array) and not x.is_fully_addressable:
        return jax.lax.with_sharding_constraint(x, replicate_sharding())
    return x


# For model states using frameworks that
def _unpack_variables(state_dict, obj):
    if "variables_structure" in obj:
        variables_flat, _ = jax.tree.flatten(state_dict["variables"])
        variables = jax.tree.unflatten(obj["variables_structure"], variables_flat)
        del obj["variables_structure"], variables_flat
    else:
        variables = state_dict["variables"]
    return variables


def serialize_mcstate(
    state: MCState,
) -> dict:
    asset_manager = current_context().asset_manager

    state_dict = serialization.to_state_dict(state)
    state_dict = jax.tree.map(_replicate, state_dict)
    variables_structure = jax.tree.structure(state.variables)
    asset_manager.write_msgpack("state.msgpack", state_dict)

    return {
        "sampler": state.sampler,
        "model": state._model,  # write the bare model
        "variables_structure": variables_structure,
    }


def deserialize_vstate(
    cls,
    obj,
) -> MCState:
    asset_manager = current_context().asset_manager

    state_dict = asset_manager.read_msgpack("state.msgpack")
    variables = _unpack_variables(state_dict, obj)
    state = cls(**obj, variables=variables)
    state = serialization.from_state_dict(state, state_dict)

    return state


register_serialization(MCState, serialize_mcstate, partial(deserialize_vstate, MCState))


def serialize_mcmixedstate(state: MCMixedState) -> dict:

    asset_manager = current_context().asset_manager

    state_dict = serialization.to_state_dict(state)
    state_dict = jax.tree.map(_replicate, state_dict)
    asset_manager.write_msgpack("state.msgpack", state_dict)

    return {
        "sampler": state.sampler,
        "sampler_diag": state.diagonal.sampler,
        "model": state._model,  # write the bare model
    }


def deserialize_mcmixedstate(obj) -> MCMixedState:
    asset_manager = current_context().asset_manager

    state_dict = asset_manager.read_msgpack("state.msgpack")
    variables = _unpack_variables(state_dict, obj)
    state = MCMixedState(**obj, variables=variables)
    state = serialization.from_state_dict(state, state_dict)
    return state


register_serialization(MCMixedState, serialize_mcmixedstate, deserialize_mcmixedstate)


def serialize_fullsumstate(state: FullSumState, *, mixed_state: bool = False) -> dict:
    asset_manager = current_context().asset_manager

    state_dict = serialization.to_state_dict(state)
    state_dict = jax.tree.map(_replicate, state_dict)
    asset_manager.write_msgpack("state.msgpack", state_dict)

    if not mixed_state:
        hilbert = state.hilbert
    else:
        hilbert = state.hilbert.physical

    return {
        "hilbert": hilbert,
        "model": state._model,  # write the bare model
    }


register_serialization(
    FullSumState, serialize_fullsumstate, partial(deserialize_vstate, FullSumState)
)



from nqxpack._src.lib_v1.closure import (
    register_closure_simple_serialization,
)
from nqxpack._src.lib_v1.custom_types import (
    register_serialization,
    register_automatic_serialization,
)

from nqxpack._src.lib_v1.lib import (
    serialize_object,
    deserialize_object,
)

from nqxpack._src.lib_v1.asset_lib import (
    InMemoryAssetManager,
    FolderAssetManager,
    ArchiveAssetManager,
)

from .versioninfo import VERSION as VERSION



"""Utilities for parsing and handling version strings."""

import re


def parse_version(version: str | tuple[int, int, int]) -> tuple[int, int, int]:
    """
    Parse a version string or tuple into a tuple of (major, minor, patch).

    Args:
        version: Either a version string like "1.2.3" or a tuple like (1, 2, 3)

    Returns:
        A tuple of (major, minor, patch)

    Examples:
        >>> parse_version("1.2.3")
        (1, 2, 3)
        >>> parse_version("1.2")
        (1, 2, 0)
        >>> parse_version("1.2.3rc1")
        (1, 2, 3)
        >>> parse_version("")
        (0, 0, 0)
        >>> parse_version((1, 2, 3))
        (1, 2, 3)
    """
    if isinstance(version, tuple):
        return version

    if isinstance(version, str):
        if version == "":
            return (0, 0, 0)
        parts = version.split(".")
        # Handle versions with fewer than 3 parts
        while len(parts) < 3:
            parts.append("0")
        try:
            return tuple(int(p) for p in parts[:3])
        except ValueError:
            # If parsing fails (e.g., "1.2.3rc1"), try to extract just the numbers
            numbers = []
            for part in parts[:3]:
                match = re.match(r"(\d+)", part)
                if match:
                    numbers.append(int(match.group(1)))
                else:
                    numbers.append(0)
            while len(numbers) < 3:
                numbers.append(0)
            return tuple(numbers)

    return (0, 0, 0)



from .common import (
    skipif_ci,
    xfailif_distributed,
    skipif_sharding,
    skipif_distributed,
    set_config,
    named_parametrize,
    hash_for_seed,
    netket_experimental_fft_autocorrelation,
)

# from .finite_diff import (
#    central_diff_grad,
#    same_derivatives
# )



import time
import socket
import sys
import os


import jax


from nqxpack._src.contextmgr import current_context

VERSION = 1.1
METADATA_VERSION = VERSION


def generate_metadata(format_version, versioninfo, extra_metadata=None):
    format_version = format_version  # ".".join(map(str, format_version))
    # versioninfo = {k: ".".join(map(str, v)) for k, v in versioninfo.items()}
    versioninfo["metadata"] = METADATA_VERSION  # ".".join(map(str, METADATA_VERSION))

    package_versions = current_context().package_versions

    if extra_metadata is None:
        extra_metadata = {}

    host_info = get_host_info()

    return {
        "format": "NetKet",
        "format_version": format_version,
        "versions": versioninfo,
        "packages": package_versions,
        "host": host_info,
        "extra": extra_metadata,
    }


def get_host_info():
    return {
        "python_version": sys.version,
        "time": time.time(),
        "hostname": socket.gethostname(),
        "n_devices": jax.device_count(),
        "uname": " ".join(os.uname()),
    }


def validate_metadata(metadata):
    if metadata["format"] != "NetKet":
        raise ValueError(f"Invalid metadata format: {metadata['format']}")



from nqxpack._src.lib_v1.custom_types import (
    register_serialization,
)

# hydra
from omegaconf import ListConfig, DictConfig, OmegaConf

register_serialization(ListConfig, OmegaConf.to_object, reconstruct_type=False)
register_serialization(DictConfig, OmegaConf.to_object, reconstruct_type=False)



import numpy as np

import jax
from jax.experimental import multihost_utils

from netket.utils import module_version

if module_version("jax") >= (0, 7, 0):
    from jax.sharding import NamedSharding
    from jax.sharding import PartitionSpec as P

from netket import config as nkconfig


def replicate_sharding():
    """
    Create a replicated sharding that works with both old and new JAX versions.
    Local copy to avoid netket_pro dependency.
    """
    # TODO: always use the NamedShardng version
    if module_version("jax") >= (0, 7, 0):
        return NamedSharding(jax.sharding.get_abstract_mesh(), P())
    else:
        from jax.sharding import PositionalSharding

        return PositionalSharding(jax.devices()).replicate()


def mode() -> str:
    """
    Returns the distributed mode used by NetKet.

    This can be one of the following: ``None``, ``"sharding"``
    """
    if nkconfig.netket_experimental_sharding:
        return "sharding"
    else:
        return None


def barrier(name: str):
    """
    Synchronizes all processes. This function ensures that all processes reach this point
    before continuing.

    Args:
        name: A unique string to identify the synchronization point.
    """
    if mode() == "sharding":
        multihost_utils.sync_global_devices(name)


def broadcast_string(s: str, root: int = 0) -> str:
    def _encode_string_to_uint64_array(s):
        """Encodes a string into a NumPy array of uint64."""
        byte_data = s.encode("utf-8")  # Convert to bytes
        padding_size = (
            8 - len(byte_data) % 8
        ) % 8  # Compute padding to make it multiple of 8
        byte_data += b"\x00" * padding_size  # Pad with null bytes
        uint64_array = np.frombuffer(byte_data, dtype=np.uint64)  # Interpret as uint64
        return uint64_array, padding_size

    def _decode_uint64_array_to_string(uint64_array, padding_size):
        """Decodes a NumPy uint64 array back to a string."""
        byte_data = uint64_array.tobytes()  # Convert back to bytes
        return (
            byte_data[:-padding_size].decode("utf-8")
            if padding_size
            else byte_data.decode("utf-8")
        )

    if jax.process_count() > 1:
        if root != 0:
            raise ValueError("Only root=0 is supported in sharding mode")

        encoded_array, pad_size = _encode_string_to_uint64_array(s)
        encoded_array = multihost_utils.broadcast_one_to_all(encoded_array)
        pad_size = multihost_utils.broadcast_one_to_all(pad_size)
        s = _decode_uint64_array_to_string(encoded_array, pad_size)

    return s


def allgather(array, *, axis: int = 0, token=None):
    """
    Gathers (unshard) a distributed (sharded) array to all processes.

    The resulting array will have the same shape as the input array except
    the first axis, which will be :ref:`netket_pro.distributed.process_count`
    times longer.

    .. note::

        An input array of shape :math:`(M, N, ...)` will lead to a gathered
        array of shape :math:`(P \times M, N, ...)`, where :math:`P` is the
        number of processes.

    .. note::

        The resulting array will be unsharded, or fully addressable locally
        and on every process.

    Args:
        array: The array to gather.

    Returns:
        A tuple of the gathered array and the token.

    """
    if axis != 0:
        raise NotImplementedError("Only axis=0 is supported for now. Open a PR.")

    if mode() == "sharding":
        sharding = replicate_sharding()
        array = jax.lax.with_sharding_constraint(array, sharding)
    else:
        pass
    return array, token



from types import FunctionType
from functools import partial

# needed to support parametric classes
from plum.parametric import type_unparametrized

import numpy as np

import dataclasses

from nqxpack._src.lib_v1.resolution import (
    _fname,
    _resolve_qualname,
    _qualname,
)
from nqxpack._src.lib_v1.custom_types import (
    TYPE_SERIALIZATION_REGISTRY,
)
from nqxpack._src.lib_v1.versioned_registry import (
    VERSIONED_DESERIALIZATION_REGISTRY,
)
from nqxpack._src.lib_v1.closure import (
    is_closure,
    CLOSURE_SERIALIZATION_REGISTRY,
)
from nqxpack._src.errors import (
    SerializationError,
)
from nqxpack._src.contextmgr import autopath, current_context

PLAIN_TYPES = (int, float, str, bool, type(None))

NUMERIC_TYPES = (int, float, complex)


@autopath
def serialize_object(obj):
    """
    Serialize an object to a JSON-serializable form similar to what hydra can process.

    To deserialize the object, use `deserialize_object`.

    This also accepts an ``asset_manager`` argument that can be used to store
    large objects that should not be serialized in the JSON file. If the asset manager is
    not specified, this function might fail.

    Args:
        obj: The object to serialize.
        path: The path to the object, as a tuple of strings. This is optional, and is used
            internally to provide better error messages.
        asset_manager: An instance of an AssetManager, which is used to store large objects
    """
    if obj is None:
        return obj
    elif isinstance(obj, PLAIN_TYPES):
        return obj
    elif isinstance(obj, list):
        return [serialize_object(x, path=i) for i, x in enumerate(obj)]
    elif type(obj) is tuple:
        if len(obj) == 0:
            return {"_target_": "builtins.tuple"}
        else:
            return {
                "_target_": "builtins.tuple",
                "_args_": [
                    [
                        serialize_object(
                            x,
                            path=("_args_", i),
                        )
                        for i, x in enumerate(obj)
                    ],
                ],
            }
    elif isinstance(obj, dict):
        # It's rare (though it happnens, for example in nnx.NodeMapping) but some dicts
        # have non-string keys. We need to convert them to strings. We do this by wrapping
        if any(not isinstance(k, str) for k in obj.keys()):
            return {
                "_target_": "nqxpack._src.lib_v1.custom_types.StringKeyDict",
                "_args_": [
                    [
                        serialize_object(k, path="_args_/{i}"),
                        serialize_object(v, path=k),
                    ]
                    for i, (k, v) in enumerate(obj.items())
                ],
            }
        else:
            return {k: serialize_object(v, path=k) for k, v in obj.items()}

    # I don't think this is needed, as it's handled by serializing the type itself.
    # elif any(obj is t for t in NUMERIC_TYPES):
    #    return np.dtype(obj).name
    elif isinstance(obj, np.dtype):
        return obj.name
    elif isinstance(obj, type):
        # special case NoneType, because it does not exist in builtins but it does in types
        if obj is type(None):
            return "< types.NoneType >"
        return "< " + _fname(obj) + " >"
    else:
        return serialize_object(serialize_custom_object(obj))


@autopath
def deserialize_object(obj):
    """
    Deserialize an object from a JSON-serializable produced by `serialize_object`.

    Args:
        obj: The object to deserialize.
        asset_manager: An instance of an AssetManager.
    """
    if obj is None:
        return obj
    elif isinstance(obj, str):
        # Check if it is a function
        if obj.startswith("<") and obj.endswith(">"):
            return _resolve_qualname(obj[1:-1].strip())
        return obj
    elif isinstance(obj, PLAIN_TYPES):
        return obj
    elif isinstance(obj, list):
        return [deserialize_object(x, path=i) for i, x in enumerate(obj)]
    elif isinstance(obj, tuple):
        return tuple(deserialize_object(x, path=i) for i, x in enumerate(obj))
    elif isinstance(obj, dict):
        obj = {k: deserialize_object(v, path=k) for k, v in obj.items()}
        if "_target_" in obj:
            obj = deserialize_custom_object(obj)
        return obj
    else:
        global_path = current_context().path
        raise NotImplementedError(
            f"Cannot deserialize object of type {type(obj)} : {obj} found at path `{global_path}`."
        )


def serialize_custom_object(obj):
    """
    Function to serialize custom objects. This is used by `serialize_object` to serialize
    custom objects that are not handled by the default serialization.

    Should not be used directly.
    """

    if hasattr(obj, "__to_json__"):
        dict_data = obj.__to_json__()
        if "_target_" not in dict_data:
            dict_data["_target_"] = _qualname(obj)
        return dict_data
    elif type_unparametrized(obj) in TYPE_SERIALIZATION_REGISTRY:
        typ = type_unparametrized(obj)
        dict_data = TYPE_SERIALIZATION_REGISTRY[typ](obj)
        return dict_data
    elif dataclasses.is_dataclass(obj):
        # no: this recurses
        # dict_data = dataclasses.asdict(obj)
        dict_data = {"_target_": _qualname(obj)}
        for field in dataclasses.fields(obj):
            dict_data[field.name] = serialize_object(
                getattr(obj, field.name), path=field.name
            )
        return dict_data
    elif is_closure(obj):
        serialization_fun = CLOSURE_SERIALIZATION_REGISTRY.get(_fname(obj), None)
        if serialization_fun is not None:
            return serialization_fun(obj)
        else:
            global_path = current_context().path
            raise NotImplementedError(
                f"""
                    Cannot serialize closure object of type `{_qualname(obj)}` : `{obj}`
                    found at path `{global_path}`.

                You should register a serialization function for this closure by adding to
                netket_project/utils/serialization/serialize_v1.py the following line:

                register_closure_simple_serialization(
                    parent_function, {obj.__name__}, original_qualname='{_qualname(obj)}'
                    )
                """
            )

    elif isinstance(obj, FunctionType):
        if obj.__name__ == obj.__qualname__:  # top-level function
            return "< " + _fname(obj) + " >"
        else:
            # staticmethod
            return "< " + _fname(obj) + " >"
    else:
        global_path = current_context().path
        raise SerializationError(_qualname(obj), obj, path=global_path)


def deserialize_custom_object(obj):
    if "_target_" in obj:
        target_str = obj["_target_"]
        del obj["_target_"]

        is_custom = target_str.startswith("#")
        if is_custom:
            target_str = target_str[1:]

        # First, try to get a versioned deserializer based on the class path string
        deserialization_fun = VERSIONED_DESERIALIZATION_REGISTRY.get(target_str)
        if deserialization_fun is None:
            # Fall back to resolving the qualname and using default deserialization
            target = _resolve_qualname(target_str)
            deserialization_fun = partial(default_deserialization, target)

        try:
            return deserialization_fun(obj)
        except Exception as err:
            global_path = current_context().path
            raise RuntimeError(
                f"""
                Impossible to reconstruct object of type `{target_str}` at {global_path}.

                The custom deserialization function was called with the following arguments,
                and failed with the error reported above.

                The argumnts where:
                {obj}
                """
            ) from err
    else:
        global_path = current_context().path
        raise ValueError(
            f"Invalid serialization format for custom object at `{global_path}`."
        )


def default_deserialization(target, obj):
    try:
        if "_args_" in obj:
            args = obj["_args_"]
            del obj["_args_"]
        else:
            args = ()
        return target(*args, **obj)
    except Exception as err:
        global_path = current_context().path
        raise ValueError(
            f"Impossible to reconstruct object at {global_path}:\n obj: {target}({args}, {obj})"
        ) from err



from typing import TYPE_CHECKING

import threading
from functools import wraps

if TYPE_CHECKING:
    from nqxpack._src.lib_v1.asset_lib import AssetManager

from nqxpack._src.version_utils import parse_version

# Thread-local storage to keep track of the context per thread
_local = threading.local()

# Dictionary holding the versions of the packages used in the serialization
# until now. We cache it to speed up serialisation, and assume that package
# versions cannot change during runtime.
if not hasattr(_local, "packages"):
    _local.packages = {}


class PackingContext:
    def __init__(self, asset_manager: "AssetManager | None" = None, metadata=None):
        if metadata is None:
            metadata = {}

        self._path_stack = []  # Stack to track traversal paths
        self._metadata = metadata  # Storage for metadata
        self._asset_manager = (
            asset_manager  # Example asset manager (could be any object)
        )

        # Parse package versions from metadata if they exist
        if "packages" in metadata:
            self._parsed_package_versions = {
                pkg: parse_version(ver) for pkg, ver in metadata["packages"].items()
            }
        else:
            self._parsed_package_versions = {}

    def enter_path(self, path):
        """Pushes a path onto the stack."""
        self._path_stack.append(path)

    def exit_path(self):
        """Pops a path from the stack."""
        if self._path_stack:
            self._path_stack.pop()

    @property
    def path(self):
        """Returns the current path (top of the stack) or None if empty."""
        path = "/".join(map(str, self._path_stack))
        return path

    @property
    def raw_path(self):
        """Returns the raw path stack."""
        return tuple(self._path_stack)

    @property
    def package_versions(self):
        return _local.packages

    @property
    def saved_file_package_versions(self):
        """Returns parsed package versions as tuples of (major, minor, patch)."""
        return self._parsed_package_versions

    def set_metadata(self, key, value):
        """Stores metadata in the context."""
        self._metadata[key] = value

    def get_metadata(self, key=None):
        """Retrieves stored metadata. If key is None, returns all metadata."""
        return self._metadata if key is None else self._metadata.get(key)

    @property
    def asset_manager(self) -> "AssetManager":
        """Retrieves the asset manager object."""
        return self._asset_manager

    def __enter__(self):
        set_context(self)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        clear_context()
        pass  # Nothing to clean up in this simple case


# Singleton access for the context manager
def current_context() -> PackingContext:
    if not hasattr(_local, "context"):
        _local.context = PackingContext()
    return _local.context


def set_context(context):
    _local.context = context


def clear_context():
    if hasattr(_local, "context"):
        del _local.context


def autopath(func):
    """Decorator to automatically manage context paths."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        __tracebackhide__ = True

        ctx = current_context()
        path = kwargs.pop("path", None)
        if path:
            ctx.enter_path(path)
        try:
            return func(*args, **kwargs)
        finally:
            if path:
                ctx.exit_path()

    return wrapper



# File containing common commands for NetKet Test infrastructure

from typing import Any
import jax

from functools import partial
import os

import pytest

import netket as nk


def _is_true(x):
    if isinstance(x, str):
        xl = x.lower()
        if xl == "1" or x == "true":
            return True
    elif x == 1:
        return True
    else:
        return False


skipif_ci = pytest.mark.skipif(
    _is_true(os.environ.get("CI", False)), reason="Test too slow/broken on CI"
)
"""Use as a decorator to mark a test to be skipped when running on CI.
For example:

Example:
>>> @skipif_ci
>>> def test_my_serial_function():
>>>     your_serial_test()

"""

xfailif_distributed = pytest.mark.xfail(
    nk.config.netket_experimental_sharding and jax.process_count() > 1,
    reason="Broken under distributed.",
)

skipif_sharding = pytest.mark.skipif(
    nk.config.netket_experimental_sharding, reason="Only run without sharding"
)

skipif_distributed = pytest.mark.skipif(
    nk.config.netket_experimental_sharding and jax.process_count() > 1,
    reason="Skip if distributed",
)


class set_config:
    """
    Temporarily changes the value of the configuration `name`.

    Example:

    >>> with set_config("netket_experimental_disable_ode_jit", True):
    >>>     run_code

    """

    def __init__(self, name: str, value: Any):
        self._name = name.upper()
        self._value = value

    def __enter__(self):
        self._orig_value = nk.config.FLAGS[self._name]
        nk.config.update(self._name, self._value)

    def __exit__(self, exc_type, exc_value, traceback):
        nk.config.update(self._name, self._orig_value)


netket_experimental_fft_autocorrelation = partial(
    set_config, "NETKET_EXPERIMENTAL_FFT_AUTOCORRELATION"
)


def hash_for_seed(obj):
    """
    Hash any object into an int that can be used in `np.random.seed`, and does not change between Python sessions.

    Args:
      obj: any object with `repr` defined to show its states.
    """

    bs = repr(obj).encode()
    out = 0
    for b in bs:
        out = (out * 256 + b) % 4294967291  # Output in [0, 2**32 - 1]
    return out


def named_parametrize(argname: str, values: list):
    param_values = [pytest.param(obj, id=f"{argname}={obj}") for obj in values]
    return pytest.mark.parametrize(argname, param_values)



from typing import Any
from collections.abc import Callable
from functools import partial, wraps

# needed to support parametric classes

import numpy as np
from typing import TypeVar
import jax.numpy as jnp


from nqxpack._src.lib_v1.asset_lib import (
    AssetManager,
)
from nqxpack._src.lib_v1.resolution import (
    _qualname,
)
from nqxpack._src.lib_v1.versioned_registry import (
    VERSIONED_DESERIALIZATION_REGISTRY,
)


TYPE_SERIALIZATION_REGISTRY = {}


T = TypeVar("T")
PathT = tuple[str, ...]
SerializationFun = Callable[[T, PathT, AssetManager], dict]
ConversionFun = Callable[[T], Any]
DeserializationFun = Callable[[dict], T]


def register_serialization(
    cls: type[T],
    serialization_fun: SerializationFun | ConversionFun,
    deserialization_fun: DeserializationFun | None = None,
    reconstruct_type: bool = True,
    override: bool = False,
    min_version: tuple[int, int, int] = (0, 0, 0),
):
    """
    Register a custom serialization function and deserialization function for a given class.

    Args:
        cls: The class to register the serialization for
        serialization_fun: A function that takes an instance of the class and returns a dictionary
            with the serialized data. If the dictionary has a key "_target_", it will be used to
            deserialize the object. If it does not have it, it will be added.
            If the function does not return a dictionary, no `_target_` entry is added.
        deserialization_fun: A function that takes the serialized dictionary and returns an instance
            of the class. If unspecified, the deserialization will use the "_target_" key to resolve
            the function to call.
        reconstruct_type: If True (default) the output of ``serialization_fun`` must be a dictionary, and
            the ``_target_`` field will be used to reconstruct the original field. If False, instead, the
            output can be any type handled by the serialisation library, and will not be reserialized. This
            can be used to serialize some types that you do not actually want to serialise, and convert them
            to default types.
        min_version: Minimum package version (inclusive) that was used to save/serialize data in this format.
            Default is (0, 0, 0). This indicates "data saved by package version >= min_version uses this format".
            When loading, the deserialization function will be selected based on the version that was used to save,
            not the current package version. The deserialization function will be registered in the versioned
            registry with this min_version.
    """
    if reconstruct_type:
        if deserialization_fun is not None:
            _target_qualname = "#" + _qualname(cls, skip_register=True)
        else:
            _target_qualname = _qualname(cls, skip_register=True)

        @wraps(serialization_fun)
        def _serialize_fun(obj):
            dict_data = serialization_fun(obj)
            assert isinstance(dict_data, dict)

            if "_target_" not in dict_data:
                dict_data["_target_"] = _target_qualname
            return dict_data

    else:
        _serialize_fun = serialization_fun

    if cls in TYPE_SERIALIZATION_REGISTRY and not override:
        raise ValueError(f"Type {cls} is already registered for serialization")

    TYPE_SERIALIZATION_REGISTRY[cls] = _serialize_fun
    if deserialization_fun is not None:
        register_deserialization(
            class_path=cls,
            deserialization_fun=deserialization_fun,
            min_version=min_version,
        )


def _simple_serialize(
    cls,
    attrs,
    attr_map,
    obj,
    array_to_list,
):
    if array_to_list:

        def _getattr(obj, attr):
            val = getattr(obj, attr)
            if (
                isinstance(val, np.ndarray)
                or isinstance(val, jnp.ndarray)
                and val.ndim <= 2
            ):
                return val.tolist()
            return val

    else:
        _getattr = getattr
    res = {k: _getattr(obj, k) for k in attrs}
    for k, v in attr_map.items():
        res[k] = _getattr(obj, v)
    res["_target_"] = _qualname(obj)
    return res


def register_deserialization(
    class_path: str | type,
    deserialization_fun: DeserializationFun,
    min_version: tuple[int, int, int] = (0, 0, 0),
):
    """
    Register a deserialization function for a deprecated or older version of a class.

    This function allows registering deserialization logic without a corresponding
    serialization function, useful for maintaining backwards compatibility with
    older serialized formats.

    Args:
        class_path: Either a fully qualified class path as a string (e.g., "package.module.OldClass")
                   or a class object which will be converted to a class path.
        deserialization_fun: Function that takes a dict and returns an instance
        min_version: Minimum package version (inclusive) that was used to save/serialize data in this format.
                    Default is (0, 0, 0). This indicates "data saved by package version >= min_version uses this deserializer".
                    When loading, this deserializer will be selected based on the version that was used to save the data.
    """
    if not isinstance(class_path, str):
        class_path = _qualname(class_path, skip_register=True)

    VERSIONED_DESERIALIZATION_REGISTRY.register(
        class_path=class_path,
        deserialization_fun=deserialization_fun,
        min_version=min_version,
    )


def register_automatic_serialization(
    cls, *attrs, array_to_list: bool = False, **attr_map
):
    """
    Register a simple serialization function for a class that just serializes the given attributes.

    Args:
        cls: The class to register the serialization for
        *attrs: The attributes to serialize and pass to the constructor directily
        **attr_map: A mapping of the attribute name to the name of the attribute in the object
        array_to_list: If True, arrays with ndim <= 1 will be converted to lists
    """
    register_serialization(
        cls,
        serialization_fun=partial(
            _simple_serialize, cls, attrs, attr_map, array_to_list=array_to_list
        ),
    )


#


def serialize_np(cls, obj):
    if obj.size < 30:
        data = obj.tolist()
    else:
        data = obj.tobytes()

    return {
        "shape": obj.shape,
        "dtype": obj.dtype.name,
        "data": data,
    }


def deserialize_np(cls, obj):
    data = obj["data"]
    if isinstance(data, (list, float, int, complex)):
        return np.array(data, dtype=obj["dtype"]).reshape(obj["shape"])
    else:
        return np.frombuffer(data, dtype=obj["dtype"]).reshape(obj["shape"])


register_serialization(
    np.ndarray, partial(serialize_np, np.ndarray), partial(deserialize_np, np.ndarray)
)


def StringKeyDict(*mapping):
    """
    Used to rebuild a dictionary with string keys from a list of tuples.
    """
    return dict(mapping)
